from pathlib import Path

import dask.dataframe
import numpy as np
import pandas as pd
import pytest

from ratschlab_common.collections import set_utils
from ratschlab_common.io.chunks import repartition_dataframes
from ratschlab_common.io.dataframe_formats import HdfDataFrameFormat, \
    ParquetDataFrameFormat

DEFAULT_KEY='ids'

df_formats = [HdfDataFrameFormat(), ParquetDataFrameFormat()]

@pytest.mark.parametrize("ids,npartitions,expected", [
    ([1,2,3], 1, [1,3]),
    ([1,2,3,4], 2, [1,3,4]),
    ([1,1,1,1,1], 2, [1]),
    ([1,1,1,1,2], 2, [1,2]),
    (["A","B","C","D","E"], 2, ["A","D", "E"]),
    ([1,2,3,4,5], 2, [1,4,5]),
    ([1,2,3,4,5,6], 2, [1,4,6]),
    ([1,2,3,4,5,6,7,8], 3, [1,4,7,8])
])
def test_compute_divisions(ids, npartitions, expected):
    np.testing.assert_array_equal(repartition_dataframes._compute_divisions(ids, npartitions),np.array(expected))

# TODO: raise error on     ([], 1, []),

# TODO: test with uniform ids!

@pytest.mark.parametrize("start_partitions,end_partitions", [
    (1, 1),
    (2, 4),
    (2, 7),
    (4, 2)
])
def test_basic_repartition_dataframes(start_partitions, end_partitions):
    ddf = dask.dataframe.from_pandas(_create_simple_df(), npartitions=start_partitions)

    assert start_partitions == ddf.npartitions

    ddf_rep = repartition_dataframes.repartition_dask_dataframe(ddf, end_partitions, DEFAULT_KEY)

    # depending on the data and the partitioning scheme, we may end up with fewer partitions
    part_diff = end_partitions - ddf_rep.npartitions
    assert  part_diff == 0 or part_diff == 1

    assert _has_disjoint_keys(ddf_rep, DEFAULT_KEY)

@pytest.mark.parametrize("start_partitions,end_partitions", [
    (1, 1),
    (2, 4),
    (4, 2)
])
@pytest.mark.parametrize("df_format", df_formats)
@pytest.mark.parametrize("key", [DEFAULT_KEY, None])
def test_repartition_directory(tmpdir, start_partitions, end_partitions, df_format, key):
    df = _create_simple_df()

    ddf = dask.dataframe.from_pandas(df, npartitions=start_partitions)

    orig_path = Path(tmpdir, 'orig')
    orig_path.mkdir(exist_ok=True)

    df_format.write_dask_df(ddf, orig_path).compute()

    dest_path = Path(tmpdir, 'dest')
    g = repartition_dataframes.repartition_directory(orig_path, dest_path,
                                                     end_partitions, key, df_format)
    g.compute()

    ddf_back = df_format.read_dask_df(dest_path)
    assert end_partitions == ddf_back.npartitions # 'nice' partitioning, so this should hold

    if key:
        assert _has_disjoint_keys(ddf_back, key)


def _has_disjoint_keys(ddf, key):
    keys_per_partition = [ddf.get_partition(i)[key].unique().compute()
                          for i in range(ddf.npartitions)]
    return set_utils.are_pairwise_disjoint(keys_per_partition)

# TODO: refactor?
def _create_simple_df(records=100, n_cols=2, key=DEFAULT_KEY):
    ids = np.random.randint(0, records / 5, records)
    # TODO: refactor: take some dataframe, append column
    cols = {key: ids}
    for c in range(n_cols):
        cols['col{}'.format(c)] = np.random.rand(records)

    df = pd.DataFrame(cols)
    return df



from pathlib import Path
import tempfile
from distributed import LocalCluster


def create_dask_cluster(cores: int, memory_per_worker: int):
    """

    :param cores:
    :param memory_per_worker: not a hard limit.
    :return:
    """

    worker_args = {'local_dir': str(Path(tempfile.gettempdir(), 'dask_dir')),
                   'memory_limit': str(memory_per_worker * 1024 ** 2)}

    return LocalCluster(cores, threads_per_worker=1, **worker_args)



import doctest  # NOQA


def are_pairwise_disjoint(it, raise_error=False):
    """
    Determines whether a collection of elements are pairwise disjoint.

    :param it: iterable of elements
    :param raise_error: if a (descriptive) error shall be raised in case the
    elements iterables are not disjoint
    :return: true if all elements are pairwise disjoint

    >>> are_pairwise_disjoint([[1,2,3], [4,5,6], [7,8]])
    True
    >>> are_pairwise_disjoint([[1,2,3], [1,5]])
    False

    """
    all_elements = set()
    for (i, l) in enumerate(it):
        s = set(l)
        if all_elements.isdisjoint(s):
            all_elements = all_elements.union(s)
        else:
            if raise_error:
                raise ValueError('Set at index {} is not disjoint with'
                                 'previous sets. Common entries are {}'.
                                 format(i, all_elements.intersection(s)))
            return False

    return True




import os

import dask
import numpy as np
import pytest

from ratschlab_common.io.chunks import chunk_indexing
from ratschlab_common.io.dataframe_formats import HdfDataFrameFormat, \
    ParquetDataFrameFormat

df_formats = [HdfDataFrameFormat(), ParquetDataFrameFormat()]

@pytest.mark.parametrize("partitions", [1,3])
@pytest.mark.parametrize("dff", df_formats)
def test_chunkfile_indexing(a_df, tmpdir, dff, partitions):
    key = 'ids'

    for i in range(partitions):
        d = a_df.copy()
        d[key] = np.random.randint(i*2, (i+1)*2, len(a_df))
        dff.write_df(d, os.path.join(tmpdir, 'part-{}.{}'.format(i, dff.format_extension())))

    with dask.config.set(scheduler=dask.local.get_sync):
        index_path = chunk_indexing.index_chunkfiles(tmpdir, key, dff)

    index = chunk_indexing.load_chunkfile_index(index_path)

    assert len(index) > 0

    for k in index.keys():
        assert index[k] == 'part-{}.{}'.format(int(int(k)/2.0), dff.format_extension())


def test_chunkfile_indexing_enforce_disjoint(a_df, tmpdir):
    key = 'ids'
    dff = df_formats[0]

    for i in range(2):
        d = a_df.copy()
        d[key] = d[key] = np.random.randint(0, 2, len(a_df))
        dff.write_df(d, os.path.join(tmpdir, 'part-{}.{}'.format(i,
                                                                 dff.format_extension())))

    with dask.config.set(scheduler=dask.local.get_sync):
        with pytest.raises(ValueError):
            chunk_indexing.index_chunkfiles(tmpdir, key, dff)




import math
import logging

from pyspark.sql import SparkSession

from ratschlab_common.db.utils import PostgresDBConnectionWrapper, \
    PostgresDBParams


class PostgresTableDumper:
    def __init__(self, params: PostgresDBParams, spark_session: SparkSession,
                 partition_size: int = 300):
        """

        :param params:
        :param spark_session:
        :param partition_size: that much worth of data in MB on postgres
        would get stored into one partition
        """

        self.spark = spark_session
        self.params = params
        self.partition_size = partition_size

    def dump_table(self, table_name, path, partition_column=None,
                   nr_partitions=None):
        """

        :param table_name:
        :param path:
        :param partition_column: integer column!
        :param nr_partitions:
        :return:
        """
        # TODO: warnings if no partitions for large tables

        # TODO: warnings if partition not indexed

        if partition_column and not nr_partitions:
            # figure out number of partitions
            with PostgresDBConnectionWrapper(self.params) as db:
                tbl_size = db.table_size(table_name)

            nr_partitions = int(math.ceil(tbl_size /
                                          (self.partition_size * 1024 ** 2.0)))

        if partition_column:
            min_val, max_val = self._compute_min_max_values(table_name,
                                                            partition_column)

            df = self.spark.read.jdbc(self.params.jdbc_database_url(),
                                      table_name,
                                      column=partition_column,
                                      lowerBound=int(min_val),
                                      upperBound=int(max_val),
                                      numPartitions=nr_partitions,
                                      properties=self.params.to_jdbc_dict())
        else:
            df = self.spark.read.jdbc(self.params.jdbc_database_url(),
                                      table_name,
                                      properties=self.params.to_jdbc_dict())

        logging.info("Dumping to %s using %s partitions", str(path),
                     nr_partitions if nr_partitions else 1)

        if partition_column:
            df = df.sortWithinPartitions(partition_column)

        df.write.parquet(str(path))

    def _compute_min_max_values(self, table_name, col_name):
        with PostgresDBConnectionWrapper(self.params) as db:
            r = db.raw_query("SELECT MIN({}) AS min, MAX({}) AS max FROM {}".
                             format(col_name, col_name, table_name)).next()

            if r['min'] is None or r['max'] is None:
                return 0, 1

            return int(r['min']), int(r['max'])



from pyspark.sql import SparkSession
from pyspark.conf import SparkConf
from pathlib import Path


def default_spark_config(cores: int, memory_per_executor: int, driver_overhead: int = 2000,
                         tmp_dir: str = '', extra_java_options: str = '', enable_arrow: bool = True, use_utc: bool= False) -> SparkConf:
    '''
    Constructs SparkConf object with sensible defaults for the ratschlab environment

    See also https://spark.apache.org/docs/latest/configuration.html for more information
    about the semantics of the configurations

    :param cores: number of executors (workers)
    :param memory_per_executor: memory per executor [MB]
    :param driver_overhead: Memory to allocate for the driver [MB], excluding exector memory
    :param tmp_dir: "scratch" space for spark, typically /tmp by default
    :param extra_java_options: extra parameters for the JVM
    :param enable_arrow: see https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#pyspark-usage-guide-for-pandas-with-apache-arrow , requires pyarrow to be installed
    :return: SparkConf instance
    '''
    driver_mem = cores * memory_per_executor + driver_overhead

    jars = [str(p) for p in Path(Path(__file__).parent.parent, "jars").glob(
        '*.jar')]
    jar_paths = ':'.join(jars)

    cfg = SparkConf()

    if tmp_dir:
        cfg.set("spark.local.dir", tmp_dir)

    java_options = str(extra_java_options)

    if use_utc:
        # avoiding trouble with JDBC and timestamps
        java_options = "-Duser.timezone=UTC " + java_options

    return (cfg.set("spark.driver.memory", "{}m".format(driver_mem)).
            set("spark.executor.memory", "{}m".format(memory_per_executor)).
            set("spark.driver.extraJavaOptions", java_options).
            set("spark.master", "local[{}]".format(cores)).
            set("spark.jars", jar_paths).
            set("spark.sql.execution.arrow.enabled", str(enable_arrow))
            )


def create_spark_session(cores: int, memory_per_executor: int) -> SparkSession:
    '''
    Creates a local spark session with a given number of executors and memory

    :param cores: number of executors (workers)
    :param memory_per_executor: memory per executor [MB]. A default overhead for the driver is added
    :return: SparkSession instance
    '''
    return create_spark_session_from_config(default_spark_config(cores, memory_per_executor))


def create_spark_session_from_config(cfg: SparkConf) -> SparkSession:
    return (SparkSession.
             builder.
             config(conf=cfg).
             getOrCreate())



import pytest
import os
import pandas.testing

from ratschlab_common.io.dataframe_formats import HdfDataFrameFormat, ParquetDataFrameFormat


@pytest.mark.parametrize("dff", [
    HdfDataFrameFormat(),
    ParquetDataFrameFormat()
])
def test_write_df(a_df, dff, tmpdir):
    f = os.path.join(tmpdir, 'test_file.{}'.format(dff.format_extension()))

    dff.write_df(a_df, f)

    mydf_back = dff.read_df(f)
    pandas.testing.assert_frame_equal(a_df, mydf_back)

@pytest.mark.parametrize("dff", [
    HdfDataFrameFormat(),
    ParquetDataFrameFormat()
])
def test_write_df_compression(a_df, tmpdir, dff):
    f_compr = os.path.join(tmpdir, 'test_file_compr.{}'.format(dff.format_extension()))
    f = os.path.join(tmpdir, 'test_file.{}'.format(dff.format_extension()))

    dff.write_df(a_df, f_compr, None)

    if isinstance(dff, HdfDataFrameFormat):
        options = {'complevel':0}
    elif isinstance(dff, ParquetDataFrameFormat):
        options = {'compression': None}
    else:
        raise TypeError("Don't know how to handle type {}.".format(type(dff)))

    dff.write_df(a_df, f, options)

    compr_size = os.path.getsize(f_compr)
    uncompr_size = os.path.getsize(f)

    assert compr_size < uncompr_size




import pytest
import pyarrow as pa
import pyarrow.parquet as pq
import pandas as pd
import numpy as np
from ratschlab_common.io.parquet_tools.parquet_reader import ParquetReader
from contextlib import redirect_stdout
import io

name = "file.parquet"
dim_0 = 12
dim_1 = 4

@pytest.fixture
def df_test(tmp_path):
    # generate df
    df = pd.DataFrame(np.random.rand(dim_0, dim_1), columns=['A', 'B', 'C', 'D'])
    table = pa.Table.from_pandas(df)
    pq.write_table(table, tmp_path / name)
    return df

def parse_stdout(output, row_dim, col_dim):
    output = output.strip().split('\t')
    columns = output[:4]
    values = output[4:]
    values = [float(v.strip()) for v in values]
    values = np.array(values).reshape((row_dim,col_dim))
    d = pd.DataFrame(values, columns=columns)
    return d

@pytest.fixture(params=[4,5,6])
def reader(tmp_path, request):
    return ParquetReader(tmp_path / name, batch_size=request.param)

def test_head(df_test, reader):
    with io.StringIO() as buf, redirect_stdout(buf):
        reader.head(n=5)
        output = buf.getvalue()
    d = parse_stdout(output, row_dim=5, col_dim=dim_1)
    assert np.array_equal(d.values, df_test.head().values)

def test_cat(df_test, reader):
    with io.StringIO() as buf, redirect_stdout(buf):
        reader.cat()
        output = buf.getvalue()
    d = parse_stdout(output, row_dim=dim_0, col_dim=dim_1)
    assert np.array_equal(d.values, df_test.values)

def test_tail(df_test, reader):
    with io.StringIO() as buf, redirect_stdout(buf):
        reader.tail(n=5)
        output = buf.getvalue()
    d = parse_stdout(output, row_dim=5, col_dim=dim_1)
    assert np.array_equal(d.values, df_test.values[-5:, :])



import pyarrow.parquet as pq

'''
This module provides tools to use commands like cat, head and tail with parquet file 
'''


class ParquetReader:
    def __init__(self, path, batch_size=10000):

        self.pq_file = pq.ParquetFile(path)
        self.batches = self.pq_file.reader.read_all().to_batches(batch_size)
        self.col_names = [i.name for i in self.pq_file.schema]
        self.col_types = [i.physical_type for i in self.pq_file.schema]
        self._max_len_col = max([len(name) for name in self.col_names])

    def head(self, n=5, header=True):
        if header:
            self._pretty_print_header()
        for i, lines in enumerate(self._read(max_x=n)):
            self.pretty_print(lines)

    def cat(self, header=True):
        if header:
            self._pretty_print_header()
        for i, lines in enumerate(self._read()):
            self.pretty_print(lines)

    def tail(self, n=5, header=True):
        if header:
            self._pretty_print_header()
        start = self.pq_file.metadata.num_rows - n
        for i, lines in enumerate(self._read(min_x=start)):
            self.pretty_print(lines)

    def schema(self):
        for col, t in zip(self.col_names, self.col_types):
            print(col + ': ' + t)
        print()

    def _pretty_print_header(self):
        print(" ", end="\t")
        for col in self.col_names:
            print(col, end="\t")
        print()

    def pretty_print(self, lines):

        for l, i in zip(lines.values, lines.index):
            #print(" " + str(i), end='\t')
            for el, col in zip(l, self.col_names):
                print(" " + str(el), end='\t')
            print()
        #print(lines.to_string(header=False, col_space=2))


    def _read(self, min_x=0, max_x=-1):

        l_c = 0
        for batch in self.batches:
            df = batch.to_pandas()
            if min_x >= len(df) + l_c:
                l_c += len(df)
                continue
            if max_x == -1:
                lines = df.iloc[max([min_x - l_c, 0]):]
            else:
                lines = df.iloc[max([min_x - l_c, 0]): max_x - l_c]
            l_c += len(df)
            yield lines
            if l_c >= max_x > -1:
                break




if __name__ == '''__main__''':
    path = "file_small.parquet"
    reader = ParquetReader(path)
    reader.head(n=7)
    reader.cat()
    reader.tail()
    reader.tail(n=8)




import os
from abc import ABCMeta, abstractmethod
from types import MappingProxyType

import dask.dataframe
import pandas as pd


class AbstractDataFrameFormat:
    """
    Provides helper functions to serialize and deserialize pandas and
    dask dataframes.

    Instances of such classes can define their own default options via the
    constructor. If none are specified, the library default are taken as
    default options. Default options can be overwritten in the calling
    functions.
    """
    __metaclass__ = ABCMeta

    def __init__(self, default_reading_options=None,
                 default_writing_options=None):
        self.default_reading_options = default_reading_options
        self.default_writing_options = default_writing_options

    @abstractmethod
    def read_df(self, file_path, options=None):
        raise NotImplementedError

    @abstractmethod
    def write_df(self, df, file_path, options=None):
        raise NotImplementedError

    @abstractmethod
    def read_dask_df(self, directory, options=None):
        raise NotImplementedError

    @abstractmethod
    def write_dask_df(self, ddf, directory, options=None,
                      compute=False):
        raise NotImplementedError

    @abstractmethod
    def format_extension(self):
        raise NotImplementedError

    @abstractmethod
    def _library_default_reading_options(self):
        raise NotImplementedError

    @abstractmethod
    def _library_default_writing_options(self):
        raise NotImplementedError

    def get_default_reading_options(self):
        """
        Default options used, if no options are passed in the read functions

        :return: dictionary with options
        """
        if self.default_reading_options:
            return self.default_reading_options
        return self._library_default_reading_options()

    def get_default_writing_options(self):
        """
        Default options used, if no options are passed in the write functions

        :return: dictionary with options
        """
        if self.default_writing_options:
            return self.default_writing_options
        return self._library_default_writing_options()

    @staticmethod
    def _get_effective_options(params, defaults):
        if not params:
            effective_params = defaults
        else:
            effective_params = dict(defaults, **params)

        return effective_params


class ParquetDataFrameFormat(AbstractDataFrameFormat):
    DEFAULT_ENGINE = 'pyarrow'

    def read_df(self, file_path, options=None):
        e = ParquetDataFrameFormat._get_effective_options(options,
                                                          self.get_default_reading_options())
        return pd.read_parquet(file_path, **e)

    def write_df(self, df, file_path, options=None):
        e = self._get_effective_options(options,
                                        self.get_default_writing_options())
        df.to_parquet(file_path, **e)

    def read_dask_df(self, directory, options=None):
        in_pattern = os.path.join(directory,
                                  '*.{}'.format(self.format_extension()))
        e = ParquetDataFrameFormat._get_effective_options(options,
                                                          self.get_default_reading_options())
        return dask.dataframe.read_parquet(str(in_pattern), **e)

    def write_dask_df(self, ddf, directory, options=None,
                      compute=False):
        e = ParquetDataFrameFormat._get_effective_options(options,
                                                          self.get_default_writing_options())
        return ddf.to_parquet(str(directory), compute=compute, **e)

    def format_extension(self):
        return "parquet"

    def _library_default_reading_options(self):
        return MappingProxyType({'engine': self.DEFAULT_ENGINE})

    def _library_default_writing_options(self):
        return MappingProxyType({'engine': self.DEFAULT_ENGINE,
                                 'compression': 'snappy'})


class HdfDataFrameFormat(AbstractDataFrameFormat):
    DEFAULT_KEY = 'data'

    def read_df(self, file_path, options=None):
        eff = HdfDataFrameFormat._get_effective_options(options,
                                                        self.get_default_reading_options())
        return pd.read_hdf(file_path, **eff)

    def write_df(self, df, file_path, options=None):
        eff = HdfDataFrameFormat._get_effective_options(options,
                                                        self.get_default_writing_options())
        df.to_hdf(file_path, **eff)

    def read_dask_df(self, directory, options=None):
        in_pattern = os.path.join(directory,
                                  '*.{}'.format(self.format_extension()))
        e = HdfDataFrameFormat._get_effective_options(options,
                                                      self.get_default_reading_options())
        return dask.dataframe.read_hdf(str(in_pattern), **e)

    def write_dask_df(self, ddf, directory, options=None,
                      compute=False):
        e = HdfDataFrameFormat._get_effective_options(options,
                                                      self.get_default_writing_options())
        out_pattern = os.path.join(directory,
                                   'part-*.{}'.format(self.format_extension()))
        return ddf.to_hdf(out_pattern, compute=compute, **e)

    def format_extension(self):
        return 'h5'

    def _library_default_reading_options(self):
        return MappingProxyType({'mode': 'r', 'key': self.DEFAULT_KEY})

    def _library_default_writing_options(self):
        return MappingProxyType({'key': self.DEFAULT_KEY, 'format': 'table',
                                 'complib': 'blosc:lz4', 'complevel': 5})



import pandas as pd
import numpy as np
from pandas.api.types import is_sparse
from scipy import sparse
import tables

'''
This module exposes two functions: to_hdf and read_hdf, which usage
is to read and write sparse Pandas DataFrames into and from hdf file.
'''


def read_hdf(fpath):
    '''
    read a hdf file into a sparse Pandas DataFrame that was previously saved with the to_hdf function of this module.

    :param fpath: location of the hdf file
    :return: an instance of Pandas DataFrame
    '''
    with tables.open_file(fpath, mode='r') as f:
        sparse_group = f.get_node("/data/sparse")
        sparse_m = _read_sparse_m_coo(sparse_group)
        if np.issubdtype(f.get_node("/col_names").dtype, np.number):
            col_names = f.get_node("/col_names")
            s_col_names = f.get_node("/data/sparse/sparse_col_names")
        else:
            col_names = [c.decode('UTF-8') for c in f.get_node("/col_names")]
            s_col_names = [c.decode('UTF-8') for c in f.get_node("/data/sparse/sparse_col_names")]
        non_sparse = pd.read_hdf(fpath, key="/data/non_sparse")
        sparse = pd.DataFrame.sparse.from_spmatrix(sparse_m, columns=s_col_names)
        ret_df = pd.concat([sparse, non_sparse], axis=1)
        ret_df = ret_df[col_names]
    return ret_df

def _read_sparse_m(group):
    attributes = []
    for attribute in ('data', 'indices', 'indptr', 'shape'):
        attributes.append(getattr(group, attribute).read())
    sparse_m = sparse.csr_matrix(tuple(attributes[:3]), shape=attributes[3])
    return sparse_m

def _read_sparse_m_coo(group):
    attributes = []
    for attribute in ('data', 'row', 'col', 'shape'):
        attributes.append(getattr(group, attribute).read())
    t = tuple([attributes[1], attributes[2]])
    sparse_m = sparse.coo_matrix(tuple([attributes[0], t]), shape=attributes[3])
    return sparse_m

def _store_sparse_m_coo(m, hdf5, group, complevel=5, complib='blosc', **kwargs):
    assert (m.__class__ == sparse.coo.coo_matrix), 'm must be a coo matrix'
    for attribute in ('data', 'row', 'col', 'shape'):
        arr = np.array(getattr(m, attribute))
        _store_array_hdf(arr, hdf5, group, attribute,
                                        complevel=complevel, complib=complib, **kwargs)
    return hdf5


def _store_sparse_m(m, hdf5, group, complevel=5, complib='blosc', **kwargs):
    assert (m.__class__ == sparse.csr.csr_matrix), 'M must be a csr matrix'
    for attribute in ('data', 'indices', 'indptr', 'shape'):
        arr = np.array(getattr(m, attribute))
        _store_array_hdf(arr, hdf5, group, attribute,
                                        complevel=complevel, complib=complib, **kwargs)
    return hdf5

def _store_array_hdf(arr, hdf5, group, name, complevel=5, complib='blosc', **kwargs):
    if arr.dtype == 'object':
        arr = arr.astype('U')
    atom = tables.Atom.from_dtype(arr.dtype)
    filters = tables.Filters(complevel=complevel, complib=complib, **kwargs)
    ds = hdf5.create_carray(group, name, atom, arr.shape, filters=filters)
    ds[:] = arr

def _store_non_sparse_df(df, path, complevel=5, complib='blosc', **kwargs):
    df.to_hdf(path, key="data/non_sparse",
                         complevel=complevel, complib=complib, **kwargs)


def to_hdf(df, path, sparse_hdf_filter_dict=None, non_sparse_hdf_filter_dict=None):
    '''
    write sparse dataframe to hdf file . The default compression level and complib are:
    complevel: 5
    complib: 'blosc'
    If you wish to change that, pass a dictionary to the function call as in the following example:
    sparse_hdf_filter_dict = {
      'complevel': 6,
      'complib': 'zlib'
    }
    non_sparse_hdf_filter_dict = {
      'complevel': 7,
      'complib': 'lzo'
    }
    to_hdf(my_df, my_path, sparse_hdf_filter_dict, non_sparse_hdf_filter_dict).

    The sparse matrix is saved in COO format
    :param df: an instance of Pandas DataFrame that you wish to save to file
    :param path: path of the
    :param sparse_hdf_filter_dict: options for hdf filters (sparse part)
    :param non_sparse_hdf_filter_dict: options for hdf filters (non sparse part)
    :return:
    '''
    kwargs_sparse = sparse_hdf_filter_dict if sparse_hdf_filter_dict is not None else {}
    kwargs_non_sparse = non_sparse_hdf_filter_dict if non_sparse_hdf_filter_dict is not None else {}
    assert isinstance(kwargs_sparse, dict) and isinstance(kwargs_non_sparse, dict)

    sparse_df, non_sparse_df = _split_sparse(df)
    sparse_m = sparse_df.sparse.to_coo()
    with tables.open_file(path, mode='w') as f:
        data_group = f.create_group("/", "data")
        data_sparse_group = f.create_group(data_group, "sparse")
        _store_array_hdf(df.columns.values, f, f.root, "col_names")
        _store_sparse_m_coo(sparse_m, f, data_sparse_group, **kwargs_sparse)
        _store_array_hdf(sparse_df.columns.values, f, data_sparse_group, "sparse_col_names")
        _store_non_sparse_df(non_sparse_df, path, **kwargs_non_sparse)


def _split_sparse(df):
    sparse_col = []
    non_sparse_col = []
    for col in df.columns:
        if is_sparse(df[col]):
            sparse_col.append(col)
        else:
            non_sparse_col.append(col)
    return df[sparse_col], df[non_sparse_col]



import logging
import sys
from pathlib import Path

import click
from pyspark.sql import SparkSession

from ratschlab_common import spark_wrapper
from ratschlab_common.db.dump_tables import PostgresTableDumper
from ratschlab_common.db.utils import PostgresDBConnectionWrapper, \
    PostgresDBParams

logging.basicConfig(level=logging.INFO)


def _row_counts_parquet(path: Path, spark: SparkSession) -> int:
    return spark.read.parquet(str(path)).count()


def row_counts_match(path: Path, table_name: str, db_wrapper: PostgresDBConnectionWrapper,
                     spark: SparkSession):
    row_counts = _row_counts_parquet(path, spark)

    approx_cnt = db_wrapper.count_rows(table_name, approx=True)

    if row_counts == approx_cnt:
        return True

    logging.warning("Approximate counts didn't match, trying exact "
                    "counts. May take a while....")
    exact_cnt = db_wrapper.count_rows(table_name, approx=False)

    if row_counts != exact_cnt:
        logging.error("Counts don't match: got %s in DB but %s in "
                      "file", row_counts, exact_cnt)

    return row_counts == exact_cnt


@click.command()
@click.argument("dest-dir", type=click.Path(writable=True))
@click.option("--db-host", type=str,
              help='Database host', default='localhost')
@click.option("--db-port", type=int,
              help='Database port', default=5432)
@click.option('--db-name', type=str)
@click.option('--db-schema', type=str, default='public')
@click.option("--db-username", type=str,
              help='Database username (password should be managed via .pgpass')
@click.option("--db-password", type=str,
              help='Database password (password should be managed via .pgpass')
@click.option("--ssl-mode", type=str,
              help='SSL Mode ("disable", "require", "verify-ca", or "verify-full")', default='disable')
@click.option("--cores", type=int, default=1, help="Number of workers to use")
@click.option("--memory-per-core", type=int, default=5000, help="Memory [MB] allocated per worker")
@click.option("--force", is_flag=True, default=False,
              help='If set, table files will be overwritten')
@click.option('--default-partition-col', type=str, default=None, help="If set, partitions table by the given column (e.g. some ID), if available. The number of partitions are computed using a heurstic.")
@click.option('--partition-col', type=(str, str), multiple=True, help="pair of table name, column name designating which column the given table should be partitioned on. By default default_partition_col column is used if provided")
@click.option('--nr-partitions', type=(str, int), multiple=True, help="pair of table name, nr partitions designating how many partitions should be used for the given table. This is only needed, if the default heuristics to calculate the number of paritions per table proves inadequate.")
def main(dest_dir, db_host, db_port, db_name, db_schema, db_username, db_password, ssl_mode,
         force, cores, memory_per_core, default_partition_col, partition_col,
         nr_partitions):
    """Dumps entire database into parquet files.

       Currently, only PostgreSQL databases are supported. Large tables can be split into separate chunks partitioned on
       a given column.
    """
    partition_col_dict = {k: v for k, v in partition_col}
    nr_partitions_dict = {k: v for k, v in nr_partitions}

    dest_dir_path = Path(dest_dir)
    dest_dir_path.mkdir(exist_ok=True, parents=True)

    db_params = PostgresDBParams(user=db_username, host=db_host, password=db_password,
                                 port=db_port, db=db_name, schema=db_schema, ssl_mode=ssl_mode)

    with PostgresDBConnectionWrapper(db_params) as db_wrapper:
        tables = db_wrapper.list_tables()

        spark_cfg = spark_wrapper.default_spark_config(cores, memory_per_core, use_utc=True)
        with spark_wrapper.create_spark_session_from_config(spark_cfg) as spark:
            dumper = PostgresTableDumper(db_params, spark)
            for t in tables:
                logging.info('Dumping table %s', t)

                tbl_path = Path(dest_dir_path, t)

                if not tbl_path.exists() and not force:
                    default_col = None

                    if default_partition_col:
                        cols = db_wrapper.list_columns(t)
                        if default_partition_col in cols:
                            default_col = default_partition_col
                        else:
                            logging.warning(
                                "Default partition column %s not found among columns [%s]",
                                default_partition_col, ','.join(cols))

                    p_col = partition_col_dict.get(t, default_col)
                    nr_part = nr_partitions_dict.get(t, None)

                    dumper.dump_table(t, tbl_path, p_col, nr_part)
                else:
                    logging.info('Path %s already exists, not dumping table %s',
                                 tbl_path, t)

                counts_match = row_counts_match(tbl_path, t, db_wrapper, spark)

                if counts_match:
                    logging.info("Counts for %s match", t)
                else:
                    logging.error("Counts for %s don't match", t)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    sys.exit(main())



import datetime
from collections import namedtuple
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import tables

import ratschlab_common.io.bigmatrix as bigmatrix
from ratschlab_common.io.bigmatrix import TablesBigMatrixWriter, \
    TablesBigMatrixReader, AxisDescription, MetadataFlavor

BigMatrixElement = namedtuple('BigMatrixElement',
                              ['matrix', 'row_desc', 'col_desc'])


def _row_desc(rows, maxid, flavor):
    ids = np.random.randint(1, maxid, rows)

    timestamps = int(
        datetime.datetime(2012, 1, 1).strftime('%s')) + np.random.randint(1,
                                                                          2 * 365 * 24 * 3600,
                                                                          rows)

    spec = AxisDescription(columns=[timestamps, ids],
                           column_names=["timestamp", "id"], indices=["id"])

    if flavor == MetadataFlavor.PY_TABLES:
        return spec

    return pd.DataFrame({n: c for c, n in zip(spec.columns,
                                                  spec.column_names)})  # .set_index("id")


def _col_desc(cols, flavor):
    names = ['mycol{}'.format(i) for i in range(cols)]
    names_encoded = np.array([n.encode() for n in names])

    if flavor == MetadataFlavor.PY_TABLES:
        return AxisDescription(columns=[names_encoded],
                               column_names=["name"], indices=[])
    return pd.DataFrame({"name": names})


def _data(rows, cols):
    mat_shape = (rows, cols)

    mat = np.random.uniform(size=mat_shape)
    mat[0, 0] = np.nan
    mat[1, 0] = np.inf
    mat[0, 1] = np.infty

    return mat


def _create_bigmatrix_elements(rows, cols, maxid=10,
                               row_desc_flavor=MetadataFlavor.PY_TABLES,
                               col_desc_flavor=MetadataFlavor.PY_TABLES):
    assert rows >= 2 and cols >= 2, 'test matrix needs to be large enough'

    mat = _data(rows, cols)
    rd = _row_desc(rows, maxid, row_desc_flavor)
    cd = _col_desc(cols, col_desc_flavor)

    return BigMatrixElement(mat, rd, cd)


def _assert_bigmatrix_equal(elems, hf):
    mat_pt = hf.get_node('/{}'.format(bigmatrix.DATA_KEY))
    np.testing.assert_array_almost_equal(mat_pt, elems.matrix)


def _assert_axis_desc(orig, path, key):
    if isinstance(orig, pd.DataFrame):
        df_file = pd.read_hdf(path, key)
        pd.testing.assert_frame_equal(orig, df_file)
    elif isinstance(orig, AxisDescription):
        with tables.open_file(str(path), 'r') as hf:
            tbl = hf.get_node("/{}".format(key))

            for c, cn in zip(orig.columns, orig.column_names):
                np.testing.assert_array_equal(c, tbl.col(cn))

            assert set(tbl.indexedcolpathnames) == set(orig.indices)
    else:
        raise ValueError("type {} not supported".format(type(orig)))


def test_create_bigmatrix(tmpdir):
    elems = _create_bigmatrix_elements(10, 15)
    chunkshape = (5, 3)

    path = Path('/tmp', 'matrix_test.h5')
    if path.exists():
        path.unlink()
    writer = TablesBigMatrixWriter()
    writer.write(path, elems.matrix, elems.row_desc, elems.col_desc,
                 chunkshape=chunkshape)

    with tables.open_file(str(path), 'r') as hf:
        assert chunkshape == hf.get_node(
            '/{}'.format(bigmatrix.DATA_KEY)).chunkshape
        _assert_bigmatrix_equal(elems, hf)

    _assert_axis_desc(elems.row_desc, path, bigmatrix.ROW_DESC_KEY)
    _assert_axis_desc(elems.col_desc, path, bigmatrix.COL_DESC_KEY)


@pytest.mark.parametrize("row_desc_flavor",
                         [MetadataFlavor.PY_TABLES, MetadataFlavor.PANDAS_DF])
def test_create_bigmatrix_appending(tmpdir, row_desc_flavor):
    elems = _create_bigmatrix_elements(150, 15, row_desc_flavor=row_desc_flavor)
    chunkshape = (5, 5)
    increment_size = 10

    path = Path(tmpdir, 'matrix_test_append.h5')
    if path.exists():
        path.unlink()

    writer = TablesBigMatrixWriter()

    boundaries = np.arange(0, elems.matrix.shape[0] + 1, increment_size)

    for start, end in zip(boundaries[0:], boundaries[1:]):
        if row_desc_flavor == MetadataFlavor.PY_TABLES:
            row_desc = AxisDescription(
                columns=[e[start:end] for e in elems.row_desc.columns],
                column_names=elems.row_desc.column_names,
                indices=elems.row_desc.indices)
        else:
            row_desc = elems.row_desc.iloc[start:end]

        writer.write_or_append(path,
                               elems.matrix[start:end, ],
                               row_desc,
                               elems.col_desc, chunkshape=chunkshape)

    with tables.open_file(str(path), 'r') as hf:
        mat_pt = hf.get_node("/{}".format(bigmatrix.DATA_KEY))
        assert chunkshape == mat_pt.chunkshape

        _assert_bigmatrix_equal(elems, hf)

    _assert_axis_desc(elems.row_desc, path, bigmatrix.ROW_DESC_KEY)
    _assert_axis_desc(elems.col_desc, path, bigmatrix.COL_DESC_KEY)


@pytest.mark.parametrize("row_desc_flavor",
                         [MetadataFlavor.PY_TABLES, MetadataFlavor.PANDAS_DF])
@pytest.mark.parametrize("col_desc_flavor",
                         [MetadataFlavor.PY_TABLES, MetadataFlavor.PANDAS_DF])
def test_reading_bigmatrix(tmpdir, row_desc_flavor, col_desc_flavor):
    elems = _create_bigmatrix_elements(100, 5, maxid=5,
                                       row_desc_flavor=row_desc_flavor,
                                       col_desc_flavor=col_desc_flavor)

    if row_desc_flavor == MetadataFlavor.PY_TABLES:
        ids = np.array(elems.row_desc.columns[1])  # TODO: don't use magic
    else:
        ids = elems.row_desc["id"]

    elems.matrix[:, 0] = ids

    path = Path(tmpdir, 'bigmatrix_test_{}{}.h5'.format(row_desc_flavor,
                                                        col_desc_flavor))

    if path.exists():
        path.unlink()
    writer = TablesBigMatrixWriter()
    writer.write(path, elems.matrix, elems.row_desc, elems.col_desc)

    with TablesBigMatrixReader(path) as reader:
        np.testing.assert_array_almost_equal(elems.matrix, reader.data)

        mat = elems.matrix

        np.testing.assert_array_almost_equal(mat[0:3],
                                             reader[0:3])

        np.testing.assert_array_almost_equal(mat[5:7, 1:3], reader[5:7, 1:3])

        np.testing.assert_array_almost_equal(mat[np.ix_((5, 7), (1, 2, 3))],
                                             reader[(5, 7), (1, 2, 3)])

        np.testing.assert_array_almost_equal(mat[(5, 7), 1:3],
                                             reader[(5, 7), 1:3])

        np.testing.assert_array_almost_equal(mat[(5, 6, 7), 1:3],
                                             reader[(5, 6, 7), 1:3])

        np.testing.assert_array_almost_equal(mat[5:7, (1, 3)],
                                             reader[5:7, (1, 3)])

        np.testing.assert_array_almost_equal(mat[:, (1, 3)],
                                             reader[:, (1, 3)])

        row_idx = reader.get_row_indices('id == 4')

        assert (reader[row_idx, 0] == 4).all()

        assert (reader[row_idx, 0:1][0] == 4).all()

        cols0 = reader.get_col_indices_by_name(['mycol0', 'mycol3', 'mycol1'])
        assert (reader[row_idx, cols0][:, 0] == 4).all()

        cols1 = reader.get_col_indices_by_name(['mycol1', 'mycol0'])
        assert (reader[row_idx, cols1][:, 1] == 4).all()

        assert len(reader[(3,4,5), (0,2,4)]) > 0


@pytest.mark.parametrize("filter_name,expected",
                         [('default', 'blosc:lz4'),  # special marker
                          (None, None),
                          ('zlib', 'zlib')
                          ])
def test_create_bigmatrix_compression_filters(tmpdir, filter_name, expected):
    elems = _create_bigmatrix_elements(150, 15)

    path = Path(tmpdir, 'matrix.h5')

    if filter_name != 'default':
        if filter_name:
            filter = tables.Filters(complevel=1, complib=filter_name)
        else:
            filter = None

        TablesBigMatrixWriter().write(path, elems.matrix, elems.row_desc,
                                      elems.col_desc, compression_filter=filter)
    else:
        TablesBigMatrixWriter().write(path, elems.matrix, elems.row_desc,
                                      elems.col_desc)

    with tables.open_file(str(path), 'r') as hf:
        mat_pt = hf.get_node("/{}".format(bigmatrix.DATA_KEY))
        assert mat_pt.filters.complib == expected

@pytest.mark.parametrize("l,expected", [
    ([], []),
    ([1], slice(1,2)),
    ([1,2], slice(1,3)),
    ([1,2,5], [1,2,5]),
    ([1,5], [1,5]),
    ([1,2,3,4], slice(1,5))
])
def test_convert_slice(l, expected):
    assert TablesBigMatrixReader._convert_to_slice(l) == expected



from ratschlab_common.scripts import bigmatrix_repack as br
from ratschlab_common.io import bigmatrix
from ratschlab_common.io.bigmatrix import TablesBigMatrixWriter
import pytest
import numpy as np
from click.testing import CliRunner
from pathlib import Path
import pandas as pd
import tables
import shutil

@pytest.fixture
def simple_bigmatrix_file(tmpdir):
    rows = 10
    cols = 10
    matrix = np.random.random(size=(rows, cols))
    col_desc = pd.DataFrame.from_dict({'name': np.arange(0, cols)})
    row_desc = pd.DataFrame.from_dict({'name': np.arange(0, rows)})

    src_path = Path(tmpdir, 'src.h5')
    TablesBigMatrixWriter().write(src_path, matrix, row_desc, col_desc)

    return src_path


def test_bigmatrix_repack_base_usage(simple_bigmatrix_file, tmpdir):
    dest_path = Path(tmpdir, 'dest.h5')

    chunkshape = (3, 3)

    runner = CliRunner()
    result = runner.invoke(br.main, [str(simple_bigmatrix_file),
                                      str(dest_path),
                            '--matrix-chunkshape', str(chunkshape)])

    assert result.exit_code == 0

    with tables.open_file(str(dest_path), 'r') as hf:
        mat = hf.get_node("/{}".format(bigmatrix.DATA_KEY))
        assert mat.chunkshape == chunkshape
        assert mat.filters == TablesBigMatrixWriter.BLOSC_FILTER

        assert hf.get_node("/{}/table".format(
            bigmatrix.ROW_DESC_KEY)).filters.complib == \
               TablesBigMatrixWriter.BLOSC_FILTER.complib

        assert hf.get_node("/{}/table".format(
            bigmatrix.COL_DESC_KEY)).filters.complib == \
               TablesBigMatrixWriter.BLOSC_FILTER.complib


def test_bigmatrix_force(simple_bigmatrix_file, tmpdir):
    dest_path = Path(tmpdir, 'dest.h5')

    shutil.copy(simple_bigmatrix_file, dest_path)

    runner = CliRunner()
    result = runner.invoke(br.main, [str(simple_bigmatrix_file),
                                     str(dest_path),
                                     '--matrix-chunkshape', 'auto'])

    assert result.exit_code == 1

    result = runner.invoke(br.main, [str(simple_bigmatrix_file),
                                     str(dest_path),
                                     '--matrix-chunkshape', 'auto', '--force'])

    assert result.exit_code == 0



import attr
import pgpasslib
import records
import sqlalchemy
from records import RecordCollection


@attr.s
class PostgresDBParams(object):
    user = attr.ib(default='postgres')
    password = attr.ib(default=None, type=str)
    host = attr.ib(default='localhost')
    db = attr.ib(default='postgres')
    port = attr.ib(default=5432)
    schema = attr.ib(default='public')

    # "disable", "require", "verify-ca" and "verify-full"
    ssl_mode = attr.ib(default='disable')

    def to_jdbc_dict(self):
        if not self.password:
            pw = self._get_passwd_from_pgpass()
        else:
            pw = self.password

        # see also https://jdbc.postgresql.org/documentation/head/connect.html
        return {'user': self.user,
                'password': pw,
                'driver': "org.postgresql.Driver",
                'sslmode': self.ssl_mode,
                'currentSchema' : self.schema
                }

    def database_url(self):
        if self.password:
            name_pwd = '{}:{}'.format(self.user, self.password)
        else:
            name_pwd = self.user  # assume password is in .pgpass

        return f"postgres://{name_pwd}@{self.host}:{self.port}/{self.db}"


    def jdbc_database_url(self):
        return "jdbc:postgresql://{}:{}/{}".format(self.host, self.port,
                                                   self.db)

    def connection(self):
        return sqlalchemy.create_engine(self.database_url())

    def _get_passwd_from_pgpass(self):
        return pgpasslib.getpass(self.host, self.port, self.db, self.user)


class PostgresDBConnectionWrapper:
    def __init__(self, params: PostgresDBParams):
        self.params = params

    def __enter__(self):
        self._db = records.Database(self.params.database_url())
        self._inspector = sqlalchemy.inspect(self._db._engine)
        return self

    def count_rows(self, table_name, approx=False):
        if not approx:
            q = "SELECT COUNT(*) AS cnt FROM {}".format(table_name)
        else:
            q = "SELECT reltuples::BIGINT AS cnt FROM pg_class WHERE relname " \
                "= '{}'".format(table_name)

        return int(self._db.query(q).next()['cnt'])

    def table_size(self, table_name):
        q = "SELECT pg_relation_size(quote_ident(table_name)) AS size " \
            "FROM information_schema.tables " \
            "WHERE table_name = '{}'".format(table_name)

        r = self._db.query(q).next()
        return int(r['size'])

    def list_tables(self):
        return self._inspector.get_table_names(schema=self.params.schema)

    def raw_query(self, q) -> RecordCollection:
        if self._db is None:
            raise RuntimeError('Not connected')
        return self._db.query(q)

    def list_columns(self, table_name):
        return {d['name'] for d in self._inspector.get_columns(
            table_name, schema=self.params.schema)}

    def close(self):
        self._db.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



from pathlib import Path

import pandas as pd
from click.testing import CliRunner

from ratschlab_common.scripts import export_db_to_files
from tests.db import data_fixtures


def test_export_db_to_files_base(simple_db, tmpdir):
    db_params, db = simple_db

    dest_path = Path(tmpdir, 'dump_dest')

    runner = CliRunner()

    result = runner.invoke(export_db_to_files.main,
                           ['--db-port', db_params.port,
                            '--db-username', db_params.user,
                            '--db-password', db_params.password,
                            '--db-name', db_params.db,
                            str(dest_path)])

    assert result.exit_code == 0

    df = pd.read_parquet(Path(dest_path, data_fixtures.simple_db_table_name))
    assert df.shape[0] == data_fixtures.df_row_nr




"""
Utility functions to repartition chunkfiles. If a key is specified,
the records are repartitioned, s.t. the records having the same key are put
to the same partition, i.e. the same file.

The repartitioning is done using dask. Note, the functions return only a
computation graph, i.e. don't actually perform the repartition yet. The caller
has then to call the compute() function with an appropriate scheduler.

The repartition functions do some computations however, in case a partition
key is given, to determine the partition boundaries, i.e. determine which key
goes into which file
"""

import math
from pathlib import Path

import numpy as np


def repartition_directory(dir_path, output_path, n_partitions, key, df_format):
    """
    Takes a directory of chunkfiles and rearranges the records in a new
    directory into a specified number of chunkfiles. If a key is given,
    records having the same value at this key are put into the same partition.

    :param dir_path: str or pathlib object for input chunkfile directory
    :param output_path: str or pathlib object for output chunkfile directory
    :param n_partitions: positive int, number of final partitions
    :param key: str or None, partition key
    :param df_format: AbstractDataFrameFormat instance used for writing the
    resulting data frame
    :return: dask computation graph
    """
    ddf = df_format.read_dask_df(dir_path)

    ddf_rep = repartition_dask_dataframe(ddf, n_partitions, key)

    dest = Path(output_path)
    dest.mkdir(exist_ok=True)
    return df_format.write_dask_df(ddf_rep, dest)


def repartition_dask_dataframe(ddf, n_partitions, key):
    """
    Repartitions a dask dataframe into a given number of partitions. If a key
    is given, records with the having the same key end up in the same partition.

    :param ddf: dask data frame
    :param n_partitions: positive int, number of final partitions
    :param key: str or None, partition key
    :return: dask computation graph
    """
    if n_partitions == 1:
        return ddf

    if key:
        divs = _get_divisions(ddf, key, n_partitions)
        ret = (ddf.set_index(key).
               repartition(list(divs), force=True).
               reset_index())
    else:
        ret = ddf.repartition(npartitions=n_partitions)

    return ret


def _get_ids(ddf, key):
    return ddf[key].unique().compute()


def _compute_divisions(ids, parts):
    sorted_ids = np.sort(ids)
    step = math.ceil(len(sorted_ids) / parts)
    divs = sorted_ids[::step]
    max_elem = sorted_ids[-1]
    return np.unique(np.append(divs, max_elem))


def _get_divisions(ddf, key, n_divisions):
    ids = _get_ids(ddf, key)
    divs = _compute_divisions(ids, n_divisions)
    return divs



from ratschlab_common.io.parquet_tools.parquet_reader import ParquetReader
import sys
import click

'''
Examples: 
 pqt_tool tail path_to_file -n 7
 pqt_tool cat path_to_file
 pqt_tool head path_to_file -n 5 -H
 pqt_tool schema path_to_file 
'''

@click.group()
def pq_tool():
    pass

@pq_tool.command()
@click.argument('filepath', type=click.Path(exists=True))
@click.option('-n', type=click.INT, default=5)
@click.option('-H', is_flag=True)
def tail(filepath, n, h):
    reader = ParquetReader(filepath)
    reader.tail(n=n, header=h)

@pq_tool.command()
@click.argument('filepath', type=click.Path(exists=True))
@click.option('-n', type=click.INT, default=5)
@click.option('-H', is_flag=True)
def head(filepath, n, h):
    reader = ParquetReader(filepath)
    reader.head(n=n, header=h)

@pq_tool.command()
@click.argument('filepath', type=click.Path(exists=True))
@click.option('-H', is_flag=True)
def cat(filepath, h):
    reader = ParquetReader(filepath)
    reader.cat(header=h)

@pq_tool.command()
@click.argument('filepath', type=click.Path(exists=True))
def schema(filepath):
    reader = ParquetReader(filepath)
    reader.schema()

if __name__ == '__main__':
    sys.exit(pq_tool())







import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal
from ratschlab_common.io.sparse_df import read_hdf, to_hdf
import random
import pytest


@pytest.fixture(scope='module', params=[([0,1], None), ([0,2], None), ([0,1],['a', 'b', 'c', 'd'])])
def sparse_df(request):
    # create sparse df
    random.seed(30)
    df = pd.DataFrame(np.random.randn(10, 4), columns=request.param[1])
    random_index = np.random.randint(0, 2, size=(10,))
    for i in request.param[0]:
        df[df.columns[i]][random_index == 0] = 0.0
        df[df.columns[i]] = pd.arrays.SparseArray(df[df.columns[i]], fill_value=0.0)
    return df

def test_round_trip(sparse_df, tmp_path):
    # create writer and save to disk
    df_w = sparse_df
    path = tmp_path / "myhdf5file.h5"
    to_hdf(df_w, path)

    # create reader and read df
    df_r = read_hdf(path)

    assert_frame_equal(df_w, df_r)

def test_column_order(sparse_df, tmp_path):
    df_w = sparse_df
    path = tmp_path / "myhdf5file.h5"
    to_hdf(df_w, path)

    # create reader and read df
    df_r = read_hdf(path)
    assert list(df_w.columns.values) == list(df_r.columns.values)

def test_options(sparse_df, tmp_path):
    df_w = sparse_df
    path = tmp_path / "myhdf5file.h5"
    to_hdf(df_w, path,
           sparse_hdf_filter_dict={'complevel': 6, 'complib': 'zlib'},
           non_sparse_hdf_filter_dict={'complevel': 7,'complib': 'lzo'}
           )

    # create reader and read df
    df_r = read_hdf(path)
    assert_frame_equal(df_w, df_r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

setup_requirements = ['pytest-runner']

with open('requirements.txt') as f:
    requirements = list(f.readlines())

extras_require = {
  'spark': ['pyspark >= 2.4'],
  'dask': ['dask[complete] >= 1.9.0', 'distributed >= 1.22']
}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

test_requirements = ['pytest']

setup(
    author="ETH Zurich, Biomedical Informatics Group",
    author_email='marc.zimmermann@inf.ethz.ch',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    description="Small library of common functionalities used in various projects in the ratschlab",
    entry_points={
        'console_scripts': [
            'bigmatrix-repack=ratschlab_common.scripts.bigmatrix_repack:main',
            'export-db-to-files=ratschlab_common.scripts.export_db_to_files:main',
            'pq-tool=ratschlab_common.io.parquet_tools.pq_tool:pq_tool'
        ],
    },
    install_requires=requirements,
    extras_require=extras_require,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='ratschlab_common',
    name='ratschlab_common',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ratschlab/ratschlab-common',
    version='0.3.0',
    zip_safe=False,
)



"""
Some initial code to create and load file indices for chunked datafiles.
The indices store in which file to find the data associated to a certain value
of some field, e.g. in which file the data of some patient with a specific
patientid.
"""

import glob
import json
import os

import dask
import dask.bag
import dask.dataframe


def index_chunkfiles(directory, key, df_format, index_file_path=None,
                     num_workers=4):
    """Indices a directory of chunk files.

    :param directory: path to directory
    :param key: field to index on (e.g. 'patientid')
    :param df_format: instance of AbstractDataFrameFormat
    :param index_file_path: path for the resulting index file. by default it is
    put in `directory`
    :param num_workers: how many subprocesses to use to index files in parallel
    """
    files = glob.glob(
        os.path.join(directory, '*.{}'.format(df_format.format_extension())))

    file_mappings_dsk = (
        dask.bag.from_sequence(files).
        map(lambda f: _get_map(f, key, df_format)).
        fold(_merge_maps)
    )

    file_mappings = file_mappings_dsk.compute(num_workers=num_workers)

    if not index_file_path:
        index_file_path = os.path.join(
            directory, "_{}_index.json".format(key))

    _write_index(index_file_path, file_mappings)

    return index_file_path


def load_chunkfile_index(index_file):
    """ Loads an index file as a dictionary from key to filename.
    Note, that the keys in the dictionary are string objects, ie. if you look
    for patientid 42, then the query is my_dict[str(42)]

    :param index_file: path to the index file
    """
    with open(index_file, 'r') as f:
        return json.load(f, encoding='utf-8')


def _get_map(file_path, key, df_format):
    df = df_format.read_df(file_path, {'columns': [key]})

    keys = df[key].unique()
    file_name = os.path.basename(file_path)
    return {str(k): file_name for k in keys}


def _merge_maps(m1, m2):
    if not m1.keys().isdisjoint(m2):
        msg = "Maps are supposed to have disjoint key sets!" \
              "got {} and {}".format(m1, m2)
        raise ValueError(msg)

    result = {}

    result.update(m1)
    result.update(m2)
    return result


def _write_index(file_name, mappings):
    with open(file_name, 'w') as f:
        json.dump(mappings, f)



from pathlib import Path

import pandas as pd
import pandas.testing
import pytest

from ratschlab_common import spark_wrapper
from ratschlab_common.db.dump_tables import PostgresTableDumper
from tests.db import data_fixtures


@pytest.fixture(scope="module")
def spark_session():
    cfg = spark_wrapper.default_spark_config(1, 500, use_utc=True)
    return spark_wrapper.create_spark_session_from_config(cfg)


@pytest.mark.parametrize("nr_partitions", [None, 1, 3])
@pytest.mark.slow
def test_dump_tables(simple_db, a_df: pd.DataFrame, tmpdir, spark_session,
                     nr_partitions):
    db_cfg, db_instance = simple_db

    out = Path(tmpdir, data_fixtures.simple_db_table_name)

    PostgresTableDumper(db_cfg, spark_session).dump_table(data_fixtures.simple_db_table_name, out,
                                                          nr_partitions=nr_partitions,
                                                          partition_column=data_fixtures.simple_db_partition_col)

    # read back
    df = pd.read_parquet(out)

    a_df_sorted = a_df.sort_values([data_fixtures.simple_db_partition_col, 'col0']).reset_index(
        drop=True)

    df_sorted = df.sort_values([data_fixtures.simple_db_partition_col, 'col0']).reset_index(
        drop=True)

    pandas.testing.assert_frame_equal(a_df_sorted, df_sorted)

    expected_partitions = nr_partitions if nr_partitions is not None else 1

    assert expected_partitions == len(list(out.glob('*.parquet')))



import sys
import click
import subprocess
from ratschlab_common.io import bigmatrix
from pathlib import Path

DEFAULT_MAT_OPT = ['--keep-source-filters']
DEFAULT_ROW_OPT = ['--keep-source-filters']
DEFAULT_COL_OPT = ['--keep-source-filters']


def _call_ptrepack(src, dest, key, opts):
    args = ['ptrepack']
    args.extend(opts)
    args.extend(["{}:/{}".format(src, key),
                 "{}:/{}".format(dest, key)])

    subprocess.call(args)


@click.command()
@click.argument("source-file", type=click.Path(exists=True))
@click.argument("dest-file", type=click.Path(exists=False))
@click.option("--matrix-chunkshape", type=str,
              help='Chunkshape of the matrix. Can be a tuple like "(10, 20)" '
                   'or "keep" or "auto"')
@click.option("--matrix-repack-options", type=str,
              help='options passed to ptrepack for the matrix node. If set, '
                   'the matrix-chunkshape option is ignored.')
@click.option("--row-meta-repack-options", type=str,
              help='options passed to ptrepack for the row metadata node.')
@click.option("--col-meta-repack-options", type=str,
              help='options passed to ptrepack for the column metadata node.')
@click.option("--force", is_flag=True, default=False,
              help='If set, destination file will be overwritten if it exists')
def main(source_file, dest_file, matrix_chunkshape, matrix_repack_options,
         row_meta_repack_options, col_meta_repack_options, force):
    """
    Repacks bigmatrix h5 files using pytables' ptrepack.

    Main use case is to experiment with different chunkshapes for the matrix.
    """

    if Path(dest_file).exists():
        if force:
            Path(dest_file).unlink()
        else:
            raise (FileExistsError("File {} already exists!".format(dest_file)))

    if matrix_repack_options:
        matrix_opts = matrix_repack_options.split(' ')
    else:
        matrix_opts = DEFAULT_MAT_OPT.copy()

        if matrix_chunkshape:
            matrix_opts.extend(['--chunkshape', matrix_chunkshape])

    _call_ptrepack(source_file, dest_file, bigmatrix.DATA_KEY, matrix_opts)

    if row_meta_repack_options:
        row_opts = row_meta_repack_options.split(' ')
    else:
        row_opts = DEFAULT_ROW_OPT

    _call_ptrepack(source_file, dest_file, bigmatrix.ROW_DESC_KEY, row_opts)

    if col_meta_repack_options:
        col_opts = col_meta_repack_options.split(' ')
    else:
        col_opts = DEFAULT_COL_OPT

    _call_ptrepack(source_file, dest_file, bigmatrix.COL_DESC_KEY, col_opts)

    return 0


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    sys.exit(main())



import pytest
import pandas as pd
from faker import Factory
import random
import testing.postgresql

from ratschlab_common.db.utils import PostgresDBParams

df_row_nr = 100

@pytest.fixture(scope='session')
def a_df():
    fake = Factory.create()
    fake.seed_instance(420)

    records = []
    for _ in range(df_row_nr):
        records.append([
            fake.name(),
            fake.email(),
            fake.date_time(),
            random.random(),
            random.randint(1, 20)
        ])

    col_names = ["col{}".format(i) for i in range(len(records[0]))]
    return pd.DataFrame.from_records(records, columns=col_names)


simple_db_table_name = 'mytable'
simple_db_partition_col = 'col4'  # integer column in 'a_df' fixture


@pytest.fixture(scope="module")
def simple_db(a_df):
    db = testing.postgresql.Postgresql()

    db_params = PostgresDBParams(port=db.settings['port'], db='test',
                                 password='test')

    conn = db_params.connection()
    a_df.to_sql(simple_db_table_name, conn, index=False)

    return db_params, db



"""
'Bigmatrix' represents an abstraction for out-of-core computation with large
matrices having some 'metadata' associated with the rows and columns. This
metadata can be for example some patient id and a timestamp for the rows
and a column name for each column, e.g. the name of features. The matrix itself
could represent the values of the features for patients at different times.

The motivation behind this abstraction is, that a dataset can almost be
represented as a matrix, i.e. has the same datatype across all columns, typically
some float or int. But a few columns represent something else, e.g. time information
or ids. The idea is to handle these metadata separately from the rest. This way
the entire dataset can be handled more efficiently.
"""

import os
from abc import ABCMeta, abstractmethod
import collections
from collections import namedtuple
from enum import Enum

import numpy as np
import pandas as pd
import tables

from ratschlab_common.io.dataframe_formats import HdfDataFrameFormat

DATA_KEY = 'data'
COL_DESC_KEY = 'col_descr'
ROW_DESC_KEY = 'row_descr'

NAME_FIELD = 'name'


class AbstractBigMatrixReader:
    __metaclass__ = ABCMeta

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def row_desc(self):
        """
        :return: pandas dataframe of row metadata
        """
        pass

    @abstractmethod
    def col_desc(self):
        """
        :return: pandas dataframe of column metadata
        """
        pass


class AbstractBigMatrixWriter:
    __metaclass__ = ABCMeta

    @abstractmethod
    def write(self, path, data, row_desc, col_desc):
        pass

    @abstractmethod
    def write_or_append(self, path, data, row_desc, col_desc):
        pass


class AxisDescription(
        namedtuple("AxisDescription", ['columns', 'column_names', 'indices'])):
    """
    Axis description, i.e. Metadata for an axis such as rows or columns.

    :param columns: list of 1-D numpy array, one element per column
    :param column_names: list of strings representing the column names of the
                         metadata table
    :param indices: list of column names which should be indexed
    """


class MetadataFlavor(Enum):
    """
    Whether the metadata on an axis is serialized as vanilla pytables or
    via pandas pytables.
    """
    PY_TABLES = 0
    PANDAS_DF = 1


class TablesBigMatrixWriter(AbstractBigMatrixWriter):
    """
    Writing a bigmatrix to HDF5 via pytables.

    Data is written, s.t. rows can be appended later, i.e. both rows in the actual
    matrix as well as rows in the row metadata table/dataframe.
    This allows also for writing a bigmatrix file incrementally.


    Important: When writing files
    incrementally, it is important to set a suitable chunkshape manually, as
    pytables will likely assume a too small chunksize based on the first write.

    Both metadata and matrix are stored using the BLOSC compressor by default.
    """

    BLOSC_FILTER = tables.Filters(complevel=5, complib='blosc:lz4')

    hdf_pandas_format = HdfDataFrameFormat()

    def write(self, path, data_matrix, row_desc, col_desc, chunkshape=None,
              compression_filter=BLOSC_FILTER):
        """
        Writes a matrix with row and column description into a hdf5 file

        :param path: path of the file
        :param data_matrix: numpy matrix
        :param row_desc: pandas dataframe or MetaDataSpec object describing the rows
        :param col_desc: pandas dataframe or MetaDataSpec object describing the columns
                         needs to contain some "name" field
        :param chunkshape: HDF chunkshape for storing the data matrix. If none is
        given it is determined by pytables.
        :param compression_filter: tables.Filters instance, by default doing blosc compression
        """
        with tables.open_file(str(path), mode='w') as hf:
            c = hf.create_earray(hf.root,
                                 DATA_KEY,
                                 atom=tables.Atom.from_dtype(
                                     data_matrix.dtype),
                                 shape=(0, data_matrix.shape[1]),
                                 filters=compression_filter,
                                 chunkshape=chunkshape)
            c.append(data_matrix)

        if isinstance(col_desc, AxisDescription):
            with tables.open_file(str(path), mode='a') as hf:
                self._write_pytables_metadata(col_desc, hf, COL_DESC_KEY,
                                              compression_filter)
        elif isinstance(col_desc, pd.DataFrame):
            self._write_pandas_metadata(col_desc, path, COL_DESC_KEY)
        else:
            raise ValueError("{} not supported".format(type(col_desc)))

        if isinstance(row_desc, AxisDescription):
            with tables.open_file(str(path), mode='a') as hf:
                self._write_pytables_metadata(row_desc, hf, ROW_DESC_KEY,
                                              compression_filter)
        elif isinstance(row_desc, pd.DataFrame):
            self._write_pandas_metadata(row_desc, path, ROW_DESC_KEY)
        else:
            raise ValueError("{} not supported".format(type(row_desc)))

    def write_or_append(self, path, data_matrix, row_desc, col_desc=None,
                        chunkshape=None, compression_filter=BLOSC_FILTER):
        """
        Writes a matrix with row and column description into a hdf5 file. If the
        file already exists, data is appended.

        :param path: path of the file
        :param data_matrix: numpy matrix
        :param row_desc: pandas dataframe or MetaDataSpec object describing the rows
        :param col_desc: pandas dataframe or MetaDataSpec object describing the columns
                         needs to contain some "name" field. If data gets appended
                         then it doesn't need to be provided.
        :param chunkshape: HDF chunkshape for storing the data matrix. If none is
        given it is determined by pytables.
        :param compression_filter: tables.Filters instance, by default doing blosc compression
        """
        if not os.path.exists(path):
            self.write(path, data_matrix, row_desc, col_desc,
                       chunkshape=chunkshape,
                       compression_filter=compression_filter)
            return

        with tables.open_file(str(path), mode='a') as hf:
            required_names = {DATA_KEY, COL_DESC_KEY, ROW_DESC_KEY}

            # first time it is called for a function.
            if all(('/{}'.format(n) in hf) for n in required_names):
                hf.get_node('/{}'.format(DATA_KEY)).append(data_matrix)

                if isinstance(row_desc, AxisDescription):
                    column_dict = {n: c for n, c in
                                   zip(row_desc.column_names, row_desc.columns)}

                    row_tbl = hf.get_node('/{}'.format(ROW_DESC_KEY))
                    lst = [column_dict[n]
                           for n in row_tbl.description._v_names]
                    row_tbl.append(lst)

                elif isinstance(row_desc, pd.DataFrame):
                    hf.close()
                    self.hdf_pandas_format.write_df(row_desc, path,
                                                    {'mode': 'a', 'append':
                                                        True, 'key': ROW_DESC_KEY})
                else:
                    raise ValueError("{} not supported".format(type(row_desc)))
            else:
                node_names = [x.name for x in hf.iter_nodes('/')]
                raise ValueError("expected nodes {} but seeing only {}".format(
                    required_names, node_names))

    def _write_pytables_metadata(self, desc, hf, key, compression_filter):
        py_desc = self._get_pytable_desc(desc.columns, desc.column_names)

        tbl = hf.create_table(hf.root, key,
                              py_desc,
                              filters=compression_filter)

        for ind_col in desc.indices:
            tbl.colinstances[ind_col].create_index()

        column_dict = {n: c for n, c in zip(desc.column_names, desc.columns)}
        lst = [column_dict[n] for n in py_desc._v_names]

        tbl.append(lst)

    def _write_pandas_metadata(self, df, path, key):
        self.hdf_pandas_format.write_df(df, path, {'key': key, 'mode': 'a'})

    def _get_pytable_desc(self, col_list, col_names):
        d = {}
        for c, cn in zip(col_list, col_names):
            col_type = tables.Col.from_dtype(c.dtype)
            d[cn] = col_type

        return tables.Description(d)


class TablesBigMatrixReader(AbstractBigMatrixReader):
    """
    Reading bigmatrix entries and metadata from HDF5 via pytables
    """

    def __init__(self, file_path):
        """
        Opens a bigmatrix file
        :param file_path: path to the bigmatrix file (pathlib Path or string)
        """
        self.file_path = file_path
        self._f = tables.open_file(str(file_path), mode='r')

        self.row_desc_node = self._f.get_node("/{}".format(ROW_DESC_KEY))
        self.col_desc_node = self._f.get_node("/{}".format(COL_DESC_KEY))
        self.data = self._f.get_node("/{}".format(DATA_KEY))

        self._row_desc_df = None
        self._col_desc_df = None

        self._row_flavor = \
            MetadataFlavor.PANDAS_DF if 'table' in self.row_desc_node else MetadataFlavor.PY_TABLES
        self._col_flavor =\
            MetadataFlavor.PANDAS_DF if 'table' in self.col_desc_node else MetadataFlavor.PY_TABLES

        self.col_names = self._read_cols()

        self._col_name_to_index = {n: i for i, n in enumerate(self.col_names)}

    def __getitem__(self, item):
        if isinstance(item, slice) or len(item) == 1:
            return self.data[item]
        elif len(item) == 2:
            row_dims, col_dims = item

            # attempt to convert to slice (e.g. 1,2,3 -> 1:3)
            if isinstance(row_dims, collections.Iterable):
                row_dims = self._convert_to_slice(row_dims)

            if isinstance(col_dims, collections.Iterable):
                col_dims = self._convert_to_slice(col_dims)

            if isinstance(row_dims, slice) or isinstance(col_dims, slice):
                return self.data[row_dims, col_dims]

            # this can be fairly inefficient...
            # https://github.com/PyTables/PyTables/issues/401
            return self.data[row_dims, :][:, col_dims]
        else:
            raise IndexError('Invalid slices {}'.format(item))

    @staticmethod
    def _convert_to_slice(iterable):
        start = None
        last = None
        for elem in iterable:
            if not start:
                start = elem
            if last and elem - last != 1:
                return iterable
            last = elem

        if start is None:
            return iterable
        return slice(start, last+1)

    def get_row_indices(self, row_query_string):
        """
        Determines row indices where the row metadata satisfies certain constraints.

        If the metadata is stored in vanilla pytables, tables.get_where_list is
        isued. In case of a pandas dataframe, pandas.DataFrame.query is used.
        Please refer to the respective documentation for details.

        :param row_query_string: query string
        :return: np.ndarray containing integer indices
        """
        if self._row_flavor == MetadataFlavor.PY_TABLES:
            return self.row_desc_node.get_where_list(row_query_string)

        return self.row_desc().query(row_query_string).index.values

    def get_col_indices_by_name(self, col_list):
        """
        Determines column indices of a list of column names. Assumes that the
        col metadata table/dataframe has a column 'name'

        :param col_list: list of strings representing column names
        :return: np.ndarray containing integer indices
        """
        return np.array(list(self._col_name_to_index[cn] for cn in col_list))

    def row_desc(self):
        """
        Row metadata as pandas dataframe.

        The dataframe is kept in memory.
        :return: pandas dataframe
        """
        if not isinstance(self._row_desc_df, pd.DataFrame):
            self._row_desc_df = (pd.read_hdf(self.file_path, key=ROW_DESC_KEY).
                                 reset_index(drop=True))
        return self._row_desc_df

    def col_desc(self):
        """
        Column metadata as pandas dataframe.

        The dataframe is kept in memory.

        :return: pandas dataframe
        """
        if not isinstance(self._col_desc_df, pd.DataFrame):
            self._col_desc_df = pd.read_hdf(self.file_path, key=COL_DESC_KEY)
        return self._col_desc_df

    def row_desc_as_table(self):
        """
        Row metadata as pytable
        :return: table object
        """
        return self._f.get_node("/{}".format(ROW_DESC_KEY))

    def col_desc_as_table(self):
        """
        Column metadata as pytable
        :return: table object
        """
        return self._f.get_node("/{}".format(COL_DESC_KEY))

    def close(self):
        """
        Closes file
        """
        self._f.close()

    def _read_cols(self):
        if self._col_flavor == MetadataFlavor.PY_TABLES:
            tbl = self.col_desc_as_table()

            cols_bytes = tbl.read(field=NAME_FIELD)
            return np.array([b.decode('UTF-8') for b in cols_bytes])

        return self.col_desc()[NAME_FIELD].values

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
